/*
 * $Id: pty.c,v 1.3 2020-10-08 20:14:12 phil Exp $
 * forkpty using posix_openpt (needed on Solaris 11)
 * from https://github.com/freebsd/freebsd.git
 * lib/libutil/pty.c
 * fa4144861a6583bf09d7274f521cb72df0604dcd
 * 2020-10-08
 */

/*-
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Copyright (c) 1990, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");
__SCCSID("@(#)pty.c	8.3 (Berkeley) 5/16/94");
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include "libpty.h"			/* Ubuntu has <pty.h> */

int
openpty(int *amaster, int *aslave, char *name, struct termios *termp,
    struct winsize *winp)
{
	const char *slavename;
	int master, slave;

	master = posix_openpt(O_RDWR|O_NOCTTY);
	if (master == -1)
		return (-1);

	if (grantpt(master) == -1)
		goto bad;

	if (unlockpt(master) == -1)
		goto bad;

	slavename = ptsname(master);
	if (slavename == NULL)
		goto bad;

	slave = open(slavename, O_RDWR);
	if (slave == -1)
		goto bad;

	*amaster = master;
	*aslave = slave;

	if (name)
		strcpy(name, slavename);
	if (termp)
		tcsetattr(slave, TCSAFLUSH, termp);
	if (winp)
		ioctl(slave, TIOCSWINSZ, (char *)winp);

	return (0);

bad:	close(master);
	return (-1);
}

int
forkpty(int *amaster, char *name, struct termios *termp, struct winsize *winp)
{
	int master, slave, pid;

	if (openpty(&master, &slave, name, termp, winp) == -1)
		return (-1);
	switch (pid = fork()) {
	case -1:
		(void)close(slave);
		return (-1);
	case 0:
		/*
		 * child
		 */
		(void)close(master);
		/* was login_tty: */
		(void)setsid();
#ifdef TIOCSCTTY
		if (ioctl(slave, TIOCSCTTY, NULL) < 0)
			return -1;
#else  /* TIOCSCTTY not defined */
		{
#ifdef TIOCNOTTY
			{
				int ttyfd = open("/dev/tty", O_RDWR);
				if (ttyfd >= 0) {
					(void) ioctl(ttyfd, TIOCNOTTY, NULL);
					(void) close(ttyfd);
				}
			}
#endif /* TIOCNOTTY defined */
			char *name = ttyname(slave);
			if (name) {
				if (slave != 0) (void) close(0);
				if (slave != 1) (void) close(1);
				if (slave != 2) (void) close(2);
				int fd = open(name O_RDWR);
				(void) close(fd);
				
			}
		}
#endif /* TIOCSCTTY not defined */
		dup2(slave, 0);
		dup2(slave, 1);
		dup2(slave, 2);
		if (slave > 2)
			close(slave);
		/* end was login_tty */
		return (0);
	}
	/*
	 * parent
	 */
	*amaster = master;
	(void) close(slave);
	return (pid);
}
